/* genscan2exons.c -- convert Genscan output to exons format
 *
 * The Genscan output file should have the following form:
 *
 *    GENSCAN 1.0     Date run: 31-Aug-99     Time: 07:40:46
 *
 *    Sequence addX : 100000 bp : 43.74% C+G : Isochore 2 (43.00 - 51.00 C+G%)
 *
 *    Parameter matrix: HumanIso.smat
 *
 *    Predicted genes/exons:
 *
 *    Gn.Ex Type S .Begin ...End .Len Fr Ph I/Ac Do/T CodRg P.... Tscr..
 *    ----- ---- - ------ ------ ---- -- -- ---- ---- ----- ----- ------
 *
 *     1.01 Intr +   1470   1515   46  1  1   87   97    27 0.584   0.97
 *     1.02 Intr +   4316   4370   55  2  1  103   89    47 0.802   5.28
 *       etc.
 *
 * $Revision: 1.9 $$Date: 2001/08/10 00:49:30 $
 */

#include "util.h"

#define SYNTAX  "genscan_file [-prob] > output_file"

enum { BUFSIZE = 5000, POSLEN = 12 };

static void get_args (int argc, char **argv, char **filename, bool *useprob);
static void print_exon (char *outbuf, int begin, int end, float prob,
                        bool useprob);

int main (int argc, char **argv) {
	char *filename = NULL;                  /* name of genscan file         */
	bool useprob = 0;                       /* print exon probabilities?    */
	FILE *fp;                               /* ptr to input file            */
	char *stat;                             /* read status                  */
	char buf [BUFSIZE];                     /* input buffer                 */
	char pred [BUFSIZE];                    /* prediction type field        */
	char dir [BUFSIZE];                     /* direction field              */
	int addr1, addr2;                       /* address fields               */
	float prob;                             /* probability field            */
	int gnum;                               /* current gene grouping        */
	char gdir;                              /* direction of this group      */
	int gstart, gend;                       /* endpoints for this group     */
	bool newgroup;                          /* starting a new group?        */
	char outbuf [BUFSIZE];                  /* accumulate exon output lines */

	/* get arguments */
	set_argv0 (argv[0]);
	get_args (argc, argv, &filename, &useprob);
	if (!filename)
		fatal (usage (SYNTAX));

	/* print header */
	printf ("# Exons converted from \"%s\" by %s.\n", filename, argv0);

	/* open input file; skip intro */
	if (!(fp = fopen (filename, "r")))
		fatalf ("Can't open file \"%s\".", filename);
	while ((stat = fgets (buf, BUFSIZE, fp)) && !strstr (buf, "---"))
		;
	if (!stat)
		fatal ("Can't find line with '---'.");

	/* process lines */
	gnum = 0; newgroup = 1;
	gdir = 0; gstart = 0; gend = 0;   /* pacify lint */
	while (fgets (buf, BUFSIZE, fp) && !strstr (buf, "Predicted")) {
		if (*buf == '\n') {
			if (!newgroup) {
				printf ("\n%c %d %d gene_%d\n", gdir, gstart, gend, gnum);
				printf (outbuf);
			}
			newgroup = 1;
			continue;
		}

		if (sscanf (buf+6, "%s %1[+-] %d %d", pred, dir, &addr1, &addr2) != 4)
			fatalf ("Can't parse line:\n%s", buf);
		if (strsame (pred, "Prom") || strsame (pred, "PlyA"))
			continue;    /* skip non-exons */
		if (!strsame (pred, "Init") && !strsame (pred, "Intr") &&
				!strsame (pred, "Term") && !strsame (pred, "Sngl"))
			fatalf ("Invalid prediction type:\n%s", buf);
		*dir = (*dir == '+' ? '>' : '<');
		if (addr1 > addr2) {
			swap (&addr1, &addr2);
		}
		if (!isdigit (buf[54]) || sscanf (buf+54, "%f", &prob) != 1 ||
				prob < 0.0 || prob > 1.0)
			fatalf ("Invalid exon probability:\n%s", buf);

		if (newgroup) {
			gnum++;
			gdir = *dir;
			gstart = addr1; gend = addr2;
			print_exon (outbuf, addr1, addr2, prob, useprob);
			newgroup = 0;
		}
		else {
			if (*dir != gdir)
				fatalf ("Exon direction does not match others:\n%s", buf);
			gstart = MIN (gstart, addr1);
			gend = MAX (gend, addr2);
			print_exon (NULL, addr1, addr2, prob, useprob);
		}
	}

	/* clean up */
	fclose (fp);
	return 0;
}

static void get_args (int argc, char **argv, char **filename, bool *useprob) {
	int i;

	for (i = 1; i < argc; i++) {
		if (strsame (argv[i], "-prob"))
			*useprob = 1;
		else if (!*filename)
			*filename = argv[i];
		else
			fatal (usage (SYNTAX));
	}
}

/* accumulate exon output until endpoint of gene is found */
static void print_exon (char *outbuf, int begin, int end, float prob,
                        bool useprob) {
	static char *p;
	static int space = 0;
	int n;

	if (outbuf) {                /* this works like strtok() */
		p = outbuf;
		space = BUFSIZE-1;    /* reserve last position for '\0' */
	}

	if (POSLEN * 2 + 15 > space)
		fatal ("Output buffer overflow: gene has too many exons.");

	n = (useprob ? sprintf (p, "%d %d P=%.3f\n", begin, end, prob)
	             : sprintf (p, "%d %d\n", begin, end));
	p += n;
	space -= n;
}
